
/*************************************************************************

  APU R 2.0

  crt.c

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  

#include "primelist.h"
#include "crt.h"
extern int Debug; 

extern int finitef_mult(int,int,int); 
extern int finitef_sub(int,int,int); 

/* Uses the Chinese remainder theorem to compute the number given
   by its residues mod a subset of the global list of primes, PrimeList 
   based on the algorithm given in Knuth, vol II, p. 274. 

   The computation is done lazily and it is assume that the 
   values v computed as in Knuth are stored outside the routine
   and will be passed back to the routine. This allows the computation
   to start where it left off rather than recomputing the v's again. 

   The prime_index tells which is the new residue provided for 
   the given number and computation starts over there. 

   prev is the previous partial sum computed. 
*/

BigInt
crt(Array<int> rem, Array<int> v, int prime_index, BigInt prev) 
{ 
  BigInt result; 
  int i,j; 
  int num_primes = Primes.Num_primes; 
  i = prime_index; 
  if (prime_index >= num_primes) { 
    cout << "Primes exhausted in crt()\n"; 
    return(BigInt(0)); 
  }

  if (i > 0) { 
    v[i] = finitef_sub(rem[i],v[0],Primes.primes[i]); 
    v[i] = finitef_mult(v[i],Primes.inverses(0,i),Primes.primes[i]); 
    for (j=1;j<i;j++) { 
      v[i] = finitef_sub(v[i],v[j],Primes.primes[i]); 
      v[i] = finitef_mult(v[i],Primes.inverses(j,i),Primes.primes[i]); 
    }
    if (Debug) 
      cout << prev << "\n"; 
    BigInt temp = BigInt(v[i]); 
    result = prev + (Primes.prime_prod(i-1))* temp;  
    if (Debug) 
      cout << result << "\n"; 
  } 
  else { 
    v[i] = rem[i]; 
    result = v[i]; 
  }
  return result;
}

BigInt
crt_mem(BigInt* result, Array<int> rem, Array<int> v, int prime_index, 
	BigInt prev) 
{ 
  int i,j; 
  int num_primes = Primes.Num_primes; 
  i = prime_index; 
  if (prime_index >= num_primes) { 
    cout << "Primes exhausted in crt()\n"; 
    return(BigInt(0)); 
  }

  if (i > 0) { 
    v[i] = finitef_sub(rem[i],v[0],Primes.primes[i]); 
    v[i] = finitef_mult(v[i],Primes.inverses(0,i),Primes.primes[i]); 
    for (j=1;j<i;j++) { 
      v[i] = finitef_sub(v[i],v[j],Primes.primes[i]); 
      v[i] = finitef_mult(v[i],Primes.inverses(j,i),Primes.primes[i]); 
    }
    if (Debug) 
      cout << prev << "\n"; 
    BigInt temp = BigInt(v[i]); 
    *result = prev + (Primes.prime_prod(i-1))* temp;  
    if (Debug) 
      cout << *result << "\n"; 
  } 
  else { 
    v[i] = rem[i]; 
    *result = v[i]; 
  }
}

BigInt
full_crt(Array<int> rem, int *checked) 
{ 
  BigInt result, partial_sum,prev; 
  int i,done; 
  Array<int> v = Array<int>(rem.size_of()); 
  done = 0; i = 0; 
  while (!done) {
    partial_sum = crt(rem, v, i, partial_sum);
    BigInt range = (Primes.prime_prod(i))/BigInt(2); 
    if (range <= partial_sum)
      result = partial_sum - Primes.prime_prod(i); 
    else result = partial_sum; 
    if (Debug) 
      cout << result << "\n"; 
    i++; 
    if (result == prev) { 
      done = 1; 
      *checked = 1; 
    }
    else if (i >= Primes.Num_primes) { 
      done = 1; 
      cout << "Prime exhausted in full_crt()\n"; 
    }
    else { 
      prev = result; 
    }
  }
  return result; 
}

BigInt multi_crt(MultiIntModp a) {
  BigInt result,prev,partial_sum; 
  int v[MULTI_NUMPRIMES]; 
  bool done = false; 
  int i = 0; 
  while (!done) { 
    partial_sum = partial_crt(a,v,i,partial_sum); 
    BigInt range = (Primes.prime_prod(i))/BigInt(2); 
    if (range <= partial_sum)
      result = partial_sum - Primes.prime_prod(i); 
    else result = partial_sum; 
    i++; 
    if (Debug) 
      cout << result << "\n"; 
    if (result == prev) done = true; 
    else if (i >= MULTI_NUMPRIMES) { 
      done = 1; 
      cout << "Primes exhausted in multi_crt() - increase MULTI_NUMPRIMES\n"; 
    }
    else prev = result; 
  }
}

BigInt
partial_crt(MultiIntModp rem, int* v, int prime_index, BigInt prev) 
{ 
  BigInt result; 
  int i,j; 
  i = prime_index; 
  if (prime_index >= MULTI_NUMPRIMES) { 
    cout << "Primes exhausted in partial_crt() - increase MULTI_NUMPRIMES\n"; 
    return(BigInt(0)); 
  }

  if (i > 0) { 
    v[i] = finitef_sub(rem.i_value(i),v[0],rem.prime_of(i)); 
    v[i] = modmult(v[i],Primes.inverses(0,i),rem.i_value(i)); 
    for (j=1;j<i;j++) { 
      v[i] = finitef_sub(v[i],v[j],Primes.primes[i]); 
      v[i] = modmult(v[i],Primes.inverses(j,i),rem.prime_of(i)); 
    }
    if (Debug) 
      cout << prev << "\n"; 
    BigInt temp = BigInt(v[i]); 
    result = prev + (Primes.prime_prod(i-1))* temp;  
    if (Debug) 
      cout << result << "\n"; 
  } 
  else { 
    v[i] = rem.i_value(i); 
    result = v[i]; 
  }
  return result;
}
