/*************************************************************************

  APU R 2.0

  mixvol_edges.h

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/

#if !defined(MIXVOL_EDGES_HDR)
#define MIXVOL_EDGES_HDR

#include "globals.h"
#include "intpoint.h"
#include "edge.h"
#include "mixvol.h"

class MixVolEdges { 
  private : 
    // These are the members which comprise the result of applying dfs()
    // to an instance of MixVolEdges. MixedCellEdges is an edge description
    // of the mixed cells, it has arrays of edges one for each mixed cell. 
    // CellIndex gives the corresponding indices sorted by lexicographic
    // order on the edge indices of each cell. MixedVols has the corresponding
    // mixed volumes of the cells. 
    Array<EdgeArray> MixedCellEdges; 
    Array<int> CellIndex; 
    Array<int> MixedVols; 

    // These are members which are used to determine the above cell structure.
    // AllEdges contains the list of all edges computed at each level of
    // the depth first search. AllCards has the corresp. cardinalities
    Array<EdgeMatrix> AllEdges; 
	   // 3-dim array, indexed by (depth, polytope, edge)
    int** AllCards;   // Cardinalities of edge sets, (depth,polytope)
    Edge** AllTestTuple; // All test edges  (depth,edge)
    int** AllTestInds; // All test indices (depth,polytope)
    int** Path; // Path[depth] = [x,y] x = polytope, y = edge no. 
    Array<MonomArray> sites; 
    int dim; 
    int cell_count; 

    // These members are used to reduce memory overhead by pre-allocating
    // them once and reusing them at each depth level of dfs().  
    double** simp_mat; // matrix used for simplex computations. 
    double* Midpoint; 
    int* pos; 
  public : 

    // Default constructor - everything 0. 
    MixVolEdges(); 
    // Constructor which initializes sizes of the various members.  
    MixVolEdges(int,int*,double**,Array<MonomArray>,bool); 

    // Various routines which are used to set up the mixed cell structure. 
    Array<EdgeMatrix> all_edges(); 
    void dfs(int,int*,int,bool); 
    int filter(int,int); 
    int good_any_tuple(int); 
    void insert_cell(Edge*,int*,int,int);
    friend int mixvol(List<Molynomial>,bool=false); 
    friend int mixvol(Array<Molynomial>,bool=false); 
};

inline MixVolEdges::MixVolEdges() { 
  AllCards = 0; 
  AllTestTuple = 0; 
  AllTestInds = 0; 
  Path = 0; 
  dim = 0; 
  cell_count = 0; 
  simp_mat = 0; 
  Midpoint = 0; 
  pos = 0; 
}

inline MixVolEdges::MixVolEdges(int num_polys,int* num_sites, double** mat,
Array<MonomArray> hull_pts, bool insert) { 
  AllEdges = Array<EdgeMatrix>(num_polys); // max depth = num_polys
  AllCards = new (int*)[num_polys]; 
  AllTestTuple = new (Edge*) [num_polys]; 
  AllTestInds = new (int*)[num_polys]; 
  Path = new (int*)[num_polys]; 
  sites = hull_pts; 
  dim = size_of(hull_pts[0][0]); 
  Midpoint = new double[dim+2]; 
  int mixedcell_bound = 0; 
  for (int i=0;i<num_polys;i++) { 
    AllEdges[i] = Array<EdgeArray>(num_polys); 
    AllCards[i] = new int[num_polys]; 
    AllTestInds[i] = new int[num_polys]; 
    Path[i] = new int[2]; 
    for (int j=0;j<num_polys;j++) { 
      AllEdges[i][j] = Array<Edge>(num_sites[j]*num_sites[j]); 
      mixedcell_bound *= num_sites[j]*num_sites[j]; 
    }
  }
  for (int j=0;j<num_polys;j++) 
    AllTestTuple[j] = new Edge[num_sites[j]*num_sites[j]]; 
  simp_mat = mat;
  if (insert) {  // means we want full mixed cell structure
    CellIndex = Array<int>(mixedcell_bound); 
    MixedVols = Array<int>(mixedcell_bound); 
    MixedCellEdges = Array<EdgeArray>(mixedcell_bound); 
    for (int j=0;j<mixedcell_bound;j++) 
      MixedCellEdges[j] = Array<Edge>(num_polys); 
  }
}

inline Array<EdgeMatrix> MixVolEdges::all_edges() { 
  return(AllEdges); 
}

void centroid(Edge*,double*,int,int); 
void squareup(double**,int,int); 
int determinant(Edge*,int,int,double**); 

#endif
