
/*************************************************************************

  APU R 2.0

  matrix.c

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  
#include "matrix.h"
#define NRANSI
#include "nr.h"


/* The code in these routines solves linear equations using the routines
   from Numerical recipes 2nd ed. The code takes care of the fact that
   matrices and arrays are indexed from 1 to N in Num. Recipes rather 
   than from 0 to N-1 as in C. That is, a usual C matrix can be passed
   to the code and it will take care of the indexing */


/* Solves the set of N linear equations A.x = b, returning the x array
   as output. The matrix A is an integer matrix. */

Array<int>
lin_solve (Matrix<int> A,Array<int> b)
{
  int dim1 = dim1_of(A); int dim2 = dim2_of(A); 
  if (dim1 != dim2) {
    cout << "Error in lin_solve - non-square matrix\n";
  }
  else {
    Matrix<float> temp_A = convert_to_float(A); 
    Array<float> temp_b = convert_to_float(b); 
    Matrix<float> matinv = matrix_inverse(temp_A);
    Array<float> temp_result = matinv * temp_b; 
    Array<int> result = convert_to_int(temp_result);
    return(result);
  }
}

Matrix<float>
convert_to_float(Matrix<int> M) 
{
  int dim1 = dim1_of(M); int dim2 = dim2_of(M);
  Matrix<float> result = Matrix<float>(dim1,dim2);
  for (int i=0;i<dim1;i++) {
    for (int j=0;j<dim2;j++) 
      result[i][j] = (float)(M[i][j]);
  }
  return(result);
}

Array<float> 
convert_to_float(Array<int> A) 
{
  int dim = size_of(A);
  Array<float> result = Array<float>(dim);
  for (int i=0;i<dim;i++) 
    result[i] = (float) (A[i]);
  return(result);
}

Array<int> 
convert_to_int(Array<float> A) 
{
  int dim = size_of(A);
  Array<int> result = Array<int>(dim);
  for (int i=0;i<dim;i++) 
    result[i] = (int) (A[i]);
  return(result);
}

/* inverts an NxN integer matrix */
Matrix<float>
matrix_inverse(Matrix<float> A) 
{
  float d;
  int N = dim1_of(A);

  float** local = copy_matrix_to_numrec(A,N,N);
  int* indx = (int*) new int[N+1];
  float* col = (float*) new float[N+1];
  ludcmp(local,N,indx,&d);

  Matrix<float> result = Matrix<float>(N,N);

  for (int j=1;j<=N;j++) {
    for (int i=1;i<=N;i++) 
      col[i] = 0.0;
    col[j] = 1.0;
    lubksb(local,N,indx,col);
    for (int i=1;i<=N;i++) 
      result[i-1][j-1] = col[i];
  }
  return(result);
}

float**
copy_matrix_to_numrec(Matrix<float> A, int dim1, int dim2)
{
  float** result = (float**) new float*[dim1+1]; 
  for (int i=0;i<dim2+1;i++) 
    result[i] = (float*) new float[dim2+1]; 

  for (int i=1;i<dim1+1;i++) {
    for (int j=1;j<dim2+1;j++) 
      result[i][j] = (float) A[i-1][j-1];
  }
  return(result);
}

/* lubksb() and ludcmp() are taken from the Numerical Recipes code */  
/* (C) Copr. 1986-92 Numerical Recipes Software H|2$.)#1,31.. */

/*
void 
lubksb(float **a, int n, int *indx, float b[])
{
	int i,ii=0,ip,j;
	float sum;

	for (i=1;i<=n;i++) {
		ip=indx[i];
		sum=b[ip];
		b[ip]=b[i];
		if (ii)
			for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
		else if (sum) ii=i;
		b[i]=sum;
	}
	for (i=n;i>=1;i--) {
		sum=b[i];
		for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
		b[i]=sum/a[i][i];
	}
}

*/
/* (C) Copr. 1986-92 Numerical Recipes Software H|2$.)#1,31.. */

/*
#include <math.h>
#define NRANSI
#include "nrutil.h"
#define TINY 1.0e-20;

void 
ludcmp(float **a, int n, int *indx, float *d)
{
	int i,imax,j,k;
	float big,dum,sum,temp;
	float *vv;

	vv=vector(1,n);
	*d=1.0;
	for (i=1;i<=n;i++) {
		big=0.0;
		for (j=1;j<=n;j++)
			if ((temp=fabs(a[i][j])) > big) big=temp;
		if (big == 0.0) nrerror("Singular matrix in routine ludcmp");
		vv[i]=1.0/big;
	}
	for (j=1;j<=n;j++) {
		for (i=1;i<j;i++) {
			sum=a[i][j];
			for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
		}
		big=0.0;
		for (i=j;i<=n;i++) {
			sum=a[i][j];
			for (k=1;k<j;k++)
				sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
			if ( (dum=vv[i]*fabs(sum)) >= big) {
				big=dum;
				imax=i;
			}
		}
		if (j != imax) {
			for (k=1;k<=n;k++) {
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			*d = -(*d);
			vv[imax]=vv[j];
		}
		indx[j]=imax;
		if (a[j][j] == 0.0) a[j][j]=TINY;
		if (j != n) {
			dum=1.0/(a[j][j]);
			for (i=j+1;i<=n;i++) a[i][j] *= dum;
		}
	}
	free_vector(vv,1,n);
}
#undef TINY
#undef NRANSI
*/

#include "polynomial.h"
extern int Debug;

Polynomial
det (const Matrix<Polynomial>& M) 
{
  int m = dim1_of(M); int n = dim2_of(M);

  Matrix<Polynomial> A = Matrix<Polynomial>(m,n); // local copy of M; 
  for (int i=0;i<m;i++) 
    for (int j=0;j<n;j++) 
      A[i][j] = M[i][j]; 
  if (Debug) 
    cout << A << "\n"; 
  int sign = 1;
  Polynomial divisor = Polynomial(1);
  int r = 0; 
  for (int k=0;k<n;k++) {
    if (r < m) {
      int p = r; 
      while ((p<m) && (is_zeropoly(A[p][k]))) { // Find non-zero pivot.
	p++; 
      }
      if (p < m) {
        if  (p != r) { // Pivot is in row p, so switch rows p and r.
	  for (int j = k;j<n;j++) {
	    Polynomial temp = A[p][j];
	    A[p][j] = A[r][j]; 
	    A[r][j] = temp;
	  }
	  sign = -sign; 
        }
	for (int i=r+1;i<m;i++) {
	  for (int j=k+1;j<n;j++) {
	    if (Debug) {
	      cout << A[r][k] << "\n"; 
	      cout << A[i][j] << "\n"; 
	      cout << A[r][j] << "\n"; 
	      cout << A[i][k] << "\n"; 
	      cout << divisor << "\n";
	      cout << A[i][j] << "\n"; 
	    }
	    A[i][j] = (A[r][k]*A[i][j] - A[r][j]*A[i][k])/divisor;
	  }
	  A[i][k] = Polynomial::zero_poly;
	}
        divisor = A[r][k]; r++;
      }
    }
    else k = n; 
  }
  Polynomial result;
  if (r==m) result = Polynomial(sign)*divisor; 
  else result = Polynomial::zero_poly;
  return result;
}

extern int Count; 

Matrix<Polynomial>
gauss_elim (const Matrix<Polynomial>& M) 
{
  int m = dim1_of(M); int n = dim2_of(M);

  Matrix<Polynomial> A = Matrix<Polynomial>(m,n); // local copy of M; 
  for (int i=0;i<m;i++) 
    for (int j=0;j<n;j++) 
      A[i][j] = M[i][j]; 
  if (Debug) 
    cout << A << "\n"; 
  int sign = 1;
  Polynomial divisor = Polynomial(1);
  int r = 0; 
  for (int k=0;k<n;k++) {
    if (r < m) {
      int p = r; 
      while ((p<m) && (is_zeropoly(A[p][k]))) { // Find non-zero pivot.
	p++; 
      }
      if (p < m) {
        if  (p != r) { // Pivot is in row p, so switch rows p and r.
	  for (int j = k;j<n;j++) {
	    Polynomial temp = A[p][j];
	    A[p][j] = A[r][j]; 
	    A[r][j] = temp;
	  }
	  sign = -sign; 
        }
	for (int i=r+1;i<m;i++) {
	  for (int j=k+1;j<n;j++) {
	    Count = Count + 4; 
	    A[i][j] = (A[r][k]*A[i][j] - A[r][j]*A[i][k])/divisor;
	  }
	  A[i][k] = Polynomial::zero_poly;
	}
        divisor = A[r][k]; r++;
      }
    }
    else k = n; 
  }
  return A;
}

