
/*************************************************************************

  APU R 2.0

  rows.c

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  
#include "det_sign.h"
#include "arith_util.h"

/* Procedure rows takes an array of decimal integers as input, representing
   the given column indices of the H matrix in the sign-determination 
   algorithm. The output is an array of row indices such that the given
   submatrix formed from the row and column indices will be non-singular
*/

Array<int>
rows (Array<int> L) 
{
  int n;

  if (size_of(L) == 0) {
    Array<int> result = Array<int>(0); 
    return result;
  }
  else if (size_of(L) == 1) {
    Array<int> result = Array<int>(1);
    result[0] = 0;
    return result;
  }
  else {
    n = 1 + floor_logb(max_of(L),3);
    Array<int> Ltemp = sort(reducemod(L,n));
    Array<int> L1 = rows(compute_L1(Ltemp));
    Array<int> L2 = add_to_array(rows(compute_L2(Ltemp)),power(3,(n-1)));
    Array<int> L3 = add_to_array(rows(compute_L3(Ltemp)),2*power(3,(n-1)));
    Array<int> result = concat(L1,L2);
    result = concat(result,L3);
    return result;
  }
}


/* L1 = {k | k = c_i mod 3^(n-1) for some i}. Here c_i are elements
  in the original column index array; we assume that these have already
  been reduced mod 3^(n-1). Also the list is assumed to be sorted so
  we need only check adjacent members. */

Array<int>
compute_L1(Array<int> L)
{
  int dim = size_of(L);
  if (dim == 0) {
    Array<int> result = Array<int>(0);
    return(result);
  }
  else {
    int count = 1;
    /* first count the number of different elt's in L1 so we can create an
       Array<int> of the size */
    int j = 0;
    for (int i=1;i<dim;i++) {
      if (L[i] != L[j]) {
	j = i;
	count++;
      }
    }
    Array<int> result = Array<int>(count);
    result[0] = L[0];
    j = 0;
    for (int i=1;i<dim;i++) {
      if (L[i] != result[j]) {
	j = j+1;
	result[j] = L[i];
      }
    }
    return(result);
  }
}


Array<int>
compute_L2(Array<int> L)
{

  int dim = size_of(L);

  if (dim < 2) {
    Array<int> result = Array<int>(0);
    return(result);
  }
  else {
    int count = 0;
    if (L[0] == L[1]) {
      count++;
    }
    for (int i=2;i<dim;i++) {
      if (L[i] == L[i-1]) {
	if (L[i-1] != L[i-2]) {
	  count++;
	}
      }
    }
    Array<int> result = Array<int>(count);
    int j = 0;
    if (L[0] == L[1]) {
      result[j] = L[0]; 
      j++;
    }
    for (int i=2;i<dim;i++) {
      if (L[i] == L[i-1]) {
	if (L[i-1] != L[i-2]) {
	  result[j] = L[i];
	  j++;
	}
      }
    }
    return(result);
  }
}

Array<int>
compute_L3(Array<int> L)
{
   int dim = size_of(L);

   if (dim < 3) {
     Array<int> result = Array<int>(0);
     return(result);
   }
   else {
     int count = 0;
     if ((L[0] == L[1]) &&
	 (L[1] == L[2]) ) {
       count++;
     }
     for (int i=3;i<dim;i++) {
       if ((L[i] == L[i-1]) &&
	   (L[i-1] == L[i-2]) ) {
	 if (L[i-2] != L[i-3])
	   count++;
       }
     }
     Array<int> result = Array<int>(count);
     int j = 0;
     if ((L[0] == L[1]) &&
	 (L[1] == L[2]) ) {
       result[j] = L[0];
       j++;
     }
     for (int i=3;i<dim;i++) {
       if ((L[i] == L[i-1]) &&
	   (L[i-1] == L[i-2]) ) {
	 if (L[i-2] != L[i-3])
	   result[j] = L[i];
	   j++;
       }
     }
     return(result);
   }
}


/* returns the (row,col)'th entry of the H matrix - note that
   row and col are start at 0 so need to add 1 when computing 
   floor_log */
int
H_elt(int row, int col)
{
  int result,m,n,i;
  Array<int> row_tri, col_tri; /* trinary versions of row and col */

  n = 1 + max(floor_logb(row,3),floor_logb(col,3));
  Matrix<int> H1 = get_H1();
  result = 1;
  row_tri = make_trinary(row,n);
  col_tri = make_trinary(col,n);
  for (i=0;i<n;i++) 
    result = result*H1[row_tri[i]][col_tri[i]];
  return(result);
}

/* returns trinary form of num where dim is the number of elements in 
   the trinary expansion of num - the least significant number is
   the 0'th element of the result array */

Array<int>
make_trinary(int num, int dim)
{
   int cur_num;  
   cur_num = num;
   Array<int> result = Array<int>(dim);
   for (int i=dim-1;i>=0;i--) {
     result[i] = cur_num/(power(3,i));
     cur_num = cur_num - power(3,i) * result[i];
   }
   return(result); 
}

int
make_decimal (Array<int> tri_num)
{
  int i,result,dim;

  dim = size_of(tri_num);
  result = 0;
  for (i=dim-1;i>=0;i--) 
    result = result + power(3,i)*tri_num[i];
  return(result);
}

/* returns new array with every element in arr reduced mod 3^(n-1) */

Array<int>
reducemod(Array<int> arr, int n) 
{
  int dim = size_of(arr);
  Array<int> result = Array<int>(dim);
  int num = power(3,(n-1)); 
  for (int i=0;i<dim;i++) 
    result[i] = arr[i] % num;
  return(result);
}
