
/*************************************************************************

  APU R 2.0

  slp_deriv.c

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  
#include "slp.h"
#include "field.h"

int binomial_coeff(int,int);
int factorial(int);
extern int Debug; 

/* Computes the partial derivative of the given slp w.r.t. the given 
   variable. The algorithm first determines if the given slp is
   actually the derivative wrt the same variable of another slp - 
   in that case, it finds the original slp node, i.e. the node which
   is at the head of the anti_deriv chain starting with the current
   slp, and returns the appropriate power wrt to which the derivative
   of the original node should be taken. Example : suppose the current
   slp is the second derivative wrt to a node representing f, i.e.
   it is f'', then, in order to prevent exponential growth of the
   slp's, we must first locate f, and take its derivative wrt x to the
   third power i.e. compute f''', given x. 

   If the variable "symbolic" is set to true the derivative computed
   will by "symbolic" in the sense that the derivatives of leaf nodes 
   which represent variables are not computed - rather a special
   node called "DIFF" is added in front of them. This allows 
   computation of the derivative just once, the particular derivative
   can then be obtained by specifying which variable w.r.t. which 
   the derivative has been taken. 
*/

Slp
diff(const Slp& slp, char* var = "",bool symbolic = false) 
{
  int var_index;
  if (!strcmp(var,"")) {
    if (symbolic) var_index = -2;
    else {
      cout << "non-symbolic diff(Slp) called with no variable\n"; 
      return(Slp::zero_slp);
    }
  }
  else var_index = (Slp::var_list.find(var))->index(); 
  Slp result = slp_partial_derivative(slp,var,var_index,symbolic);
  if (Debug) {
    Slp::Degree_eval_level++;
    cout << count_nodes(result) << "\n"; 
  }
  return(result);
}


Slp
slp_partial_derivative(const Slp& slp, char* var, int var_index,bool symbolic)
{
  int der_power = 1; 
  int temp; 
  Slp original = locate_original(slp,var_index,&der_power); 
  if (Debug) {
    Slp::Degree_eval_level++;
    temp = count_nodes(original); 
    cout << count_nodes(original) << "\n"; 
  }
  Slp result = slp_partial_derivative2(original,var,var_index,der_power,
  symbolic);
  if (Debug) {
    Slp::Degree_eval_level++;
    cout << temp << "\n"; 
    cout << count_nodes(result) << "\n"; 
  }
  return(result);
}

Slp
slp_partial_derivative2(const Slp& slp, char* var, int var_index,int power,
                        bool symbolic)
{
  int der_computed = 0;

  Slp result = find_partial_derivative(slp,var_index,power,&der_computed);
  if (!der_computed) 
    result = compute_slp_partial_derivative(slp,var,var_index,power,symbolic);
  return(result);
}

/* It's necessary to locate the original node in the case where the
   given slp is the (multiple) derivative of another node since
   taking the derivative of the original node will give a much
   smaller slp */
Slp
locate_original(const Slp& given_slp,int var_index,int* deriv_power)
{
  Slp result;

  Slp anti_deriv = anti_deriv_of(given_slp);
  if (!is_zero(anti_deriv)) {
    if (is_empty(deriv_computed_array(anti_deriv))) 
      result = given_slp;
    else if (deriv_computed_array(anti_deriv)[0] == var_index) {
    ;
    /*
      if (deriv_array(anti_deriv)[0] == given_slp) {
        *deriv_power = *deriv_power+1;
        result = locate_original(anti_deriv,var_index,deriv_power);
      }
      else result = given_slp;
      */
    }
    else result = given_slp;
  }
  else
    result = given_slp;
  return(result);
}

Slp
find_partial_derivative(const Slp& slp, int var_index, int power, 
					int* deriv_computed)
{
   Slp result;
   if (slp.deriv_done() == 1) {
     result = deriv_array(slp)[0]; 
     *deriv_computed = 1; 
   }
   else if (power == 0) {
     result = slp;
     *deriv_computed = 1;
   }
   else if (is_empty(deriv_array(slp))) {
     result = Slp::zero_slp;
     *deriv_computed = 0;
   }
   else if (is_empty(deriv_computed_array(slp))) {
     result = Slp::zero_slp;
     *deriv_computed = 0;
   }
   else if (deriv_computed_array(slp)[0] != var_index)  {
     result = Slp::zero_slp;
     *deriv_computed = 0;
   }
   else if (deriv_computed_array(slp)[0] == var_index) {
     result = find_partial_derivative(deriv_array(slp)[0],var_index,power-1,
				      deriv_computed);
   }
   return(result);
}


/* computes the derivative slp of the given slp wrt the given variable,
   i.e. the derivative has to be computed wrt to one variable only */

Slp
compute_slp_single_derivative(const Slp& slp, char* var,int var_index,
			      bool symbolic) 
{
  Slp l_der, r_der,result,temp_slp;
  Slp tempres;
  int i,j,done,curr_index;
  int coeff,coeff1,coeff2;
  int singleder,singlevar;

  result = 0;

  if (is_zero(slp)) 
    return(Slp::zero_slp);

  if ((op(slp) == ADD) || (op(slp) == SUB)) {
      l_der = slp_partial_derivative(lchild(slp),var,var_index,symbolic);
      r_der = slp_partial_derivative(rchild(slp),var,var_index,symbolic);
      if (op(slp) == ADD) {
	result = l_der + r_der;
      }
      else if (op(slp) == SUB) {
	result = l_der - r_der;
      }
      if (is_empty(deriv_array(slp))) {
	deriv_array(slp) = Array<Slp>(1);
	deriv_computed_array(slp) = Array<int>(1);
      }
      deriv_array(slp)[0] = result;
      deriv_computed_array(slp)[0] = var_index;
      slp.deriv_done(1);
      if (!is_zero(result)) 
        anti_deriv_of(result) = slp;
  }
  else if (op(slp) == MULT) {

        l_der = slp_partial_derivative(lchild(slp),var,var_index,symbolic);
        r_der = slp_partial_derivative(rchild(slp),var,var_index,symbolic);
	result = r_der * lchild(slp);
        tempres = l_der * rchild(slp);
	result = result + tempres;
        if (is_empty(deriv_array(slp))) {
	  deriv_array(slp) = Array<Slp>(1);
	  deriv_computed_array(slp) = Array<int>(1);
        }
	deriv_array(slp)[0] = result;
	deriv_computed_array(slp)[0] = var_index;
        slp.deriv_done(1);
        if (!is_zero(result))
	  anti_deriv_of(result) = slp;
  }
  else if (op(slp) == DIV) {

    l_der = slp_partial_derivative(lchild(slp),var,var_index,symbolic);
    r_der = slp_partial_derivative(rchild(slp),var,var_index,symbolic);
    tempres = rchild(slp) * l_der;
    result = lchild(slp) * r_der;
    result = tempres - result;
    tempres = rchild(slp) * rchild(slp);
    result = result / tempres;
    if (is_empty(deriv_array(slp))) {
      deriv_array(slp) = Array<Slp>(1);
      deriv_computed_array(slp) = Array<int>(1);
    }
    deriv_array(slp)[0] = result;
    deriv_computed_array(slp)[0] = var_index;
    slp.deriv_done(1);
    if (!is_zero(result))
      anti_deriv_of(result) = slp;

  }
  else if (op(slp) == EXP) {
    Field_element rval = value(rchild(slp));
    int i_temp = ((SmallInt&)(rval)).i_value();
    tempres = lchild(slp) ^ (i_temp-1);
    result = rchild(slp) * tempres;
    result = result * slp_partial_derivative(lchild(slp),var,var_index,
					     symbolic);
  }
  else if (op(slp) == NOP_VAR) 
    result = slp_single_terminal_derivative(slp,var,var_index,symbolic);
  else if (op(slp) == NOP_FE) 
    result = Slp::zero_slp;

  else if (op(slp) == DIFF) 
  // means the node represents either a 0 or 1 - deriv is zero in either case
    result = Slp::zero_slp; 
  else cout << "Error - SLP node with no legal operator\n";

  return(result);

}

Slp
slp_single_terminal_derivative(const Slp& slp, char* var, int var_index,
			       bool symbolic)
{
  Slp result;

  if (is_zero(slp)) 
    result = Slp::zero_slp; 
  else if (is_numerical(slp)) 
    result = Slp::zero_slp;
  else if (symbolic) {
    /* make Slp DIFF node with given Slp as left child */
    if (slp.deriv_done() == 1) {
      result = deriv_array(slp)[0];
    }
    else if (is_empty(deriv_array(slp))) {
      result = slp_binop(DIFF,slp,Slp::zero_slp);   
      deriv_array(slp) = Array<Slp>(1);
      deriv_computed_array(slp) = Array<int>(1);
      deriv_array(slp)[0] = result;
      deriv_computed_array(slp)[0] = -2;
      slp.deriv_done(1);
    }
    else if (deriv_computed_array(slp)[0] == -2)
      result = deriv_array(slp)[0];
    else {
      result = slp_binop(DIFF,slp,Slp::zero_slp);   
      deriv_array(slp)[0] = result;
      deriv_computed_array(slp)[0] = -2;
    }
  }
  else if (is_empty(deriv_array(slp))) { 
    deriv_array(slp) = Array<Slp>(1);
    deriv_computed_array(slp) = Array<int>(1);
    if (index(variable_of(slp)) != var_index) {
      result = Slp::zero_slp;
      deriv_array(slp)[0] = result;
      deriv_computed_array(slp)[0] = var_index;
    }
    else {
      result = Slp(1);
      deriv_array(slp)[0] = result;
      deriv_computed_array(slp)[0] = var_index;
    }
  }
  else {
    if (index(variable_of(slp)) != var_index) {
      result = Slp::zero_slp;
      deriv_array(slp)[0] = result;
      deriv_computed_array(slp)[0] = var_index;
    }
    else result = deriv_array(slp)[0]; 
  }
  return(result);
}

Slp
compute_slp_partial_derivative(const Slp& slp, char* var,int var_index, 
			       int der_power,bool symbolic)
{
  Field_element temp_fe;
  Slp lder,rder,temp_slp,temp_result,result;
  int lpower,rpower,coeff;
  if (der_power == 0) {
    result = slp;
  }
  else if (der_power = 1) {
    result = compute_slp_single_derivative(slp,var,var_index,symbolic);
  }
  else {
    if (is_zero(slp)) 
      result = Slp::zero_slp;
    else if ((op(slp) == ADD) || (op(slp) == SUB)) {
      lder = slp_partial_derivative2(lchild(slp),var,var_index,der_power,
				     symbolic);
      rder = slp_partial_derivative2(rchild(slp),var,var_index,der_power,
				     symbolic);
      if (op(slp) == ADD) 
	result = lder + rder;
      else if (op(slp) == SUB) 
	result = lder - rder;
      if (is_empty(deriv_array(slp))) {
	deriv_array(slp) = Array<Slp>(1);
	deriv_computed_array(slp) = Array<int>(1);
      }
      deriv_array(slp)[0] = result;
      deriv_computed_array(slp)[0] = var_index;
      if (!is_zero(slp)) 
        anti_deriv_of(result) = slp;
    }
    else if (op(slp) == MULT) {
     /* multiplication is handled as follows : repeated terms can
     occur under a straightforward expansion, so doing it in that
     manner will lead to very large SLP's. Instead the expansion
     is done according to the formula for multiplication. e.g.
     if h = f.g and we want to compute the twice partial wrt x,
     then d2h/dx2 = d2f/dx2.g + 2. df/dx . dg/dx + d2g/dx2.f
     which takes up less space.

     The function "update" generates the variables wrt which
     the partials have to be taken.
     */
     lpower = der_power;
     rpower = 0;
     result = Slp::zero_slp;
     while (lpower >= 0) {
       lder = slp_partial_derivative2(lchild(slp),var,var_index,lpower,
				      symbolic);
       rder = slp_partial_derivative2(rchild(slp),var,var_index,rpower,
				      symbolic);
       temp_result = lder * rder;
       coeff = binomial_coeff(der_power,lpower);
       if (coeff == 2) {
	 temp_result = temp_result + temp_result;
       }
       else if (coeff != 1) {
	 temp_slp = Slp(coeff);
	 temp_result = temp_result * temp_slp;
       }
       result = result + temp_result;
       lpower--;
       rpower++;
     }
     if (is_empty(deriv_array(slp))) {
	deriv_array(slp) = Array<Slp>(1);
	deriv_computed_array(slp) = Array<int>(1);
     }
     deriv_array(slp)[0] = result;
     deriv_computed_array(slp)[0] = var_index;
     if (!is_zero(result)) 
       anti_deriv_of(result) = slp;
    }
    else if (op(slp) == DIV) {
       lder = slp_partial_derivative2(lchild(slp),var,var_index,der_power,
				      symbolic);
       temp_result = lder * rchild(slp);
       rder = slp_partial_derivative2(rchild(slp),var,var_index,der_power,
				      symbolic);
       result = rder * lchild(slp);
       result = temp_result - result;
       temp_result = rchild(slp) * rchild(slp);
       result = result / temp_result;
       if (is_empty(deriv_array(slp))) {
	 deriv_array(slp) = Array<Slp>(1);
	 deriv_computed_array(slp) = Array<int>(1);
       }
       deriv_array(slp)[0] = result;
       deriv_computed_array(slp)[0] = var_index;
       if (!is_zero(result))
         anti_deriv_of(result) = slp;
    }
    else if (op(slp) == NOP_FE) 
       result = Slp::zero_slp; 
    else if (op(slp) == NOP_VAR) 
       result = Slp::zero_slp;  // der_power > 1
    else if (op(slp) == DIFF) 
    // means the node represents either a 0 or 1 - deriv is zero in either case
      result = Slp::zero_slp; 
    else 
      cout << "Error - slp node with invalid operator\n";
  }
  return(result);
}

int
binomial_coeff(int n,int k)
{
  int res,i;

  if (n==k)
    return(1);
  if (k==0)
    return(1);
  res = 1;
  if (k > n/2)
    k = n-k;
  i = n-k+1;
  while (i <= n) {
    res = res*i;
    i++;
  }
  res = res/factorial(k);
  return(res);
}


int
factorial(int n)
{
  int i,res;
  res = 1;
  i = 1;
  while (i <= n) {
    res = res*i;
    i++;
  }
  return(res);
}

