
/*************************************************************************

  APU R 2.0

  slp_eval.c

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  
#include "field.h"
#include "slp.h"
#include "slp_eval.h"
#include "polynomial.h"

extern int Debug; 

/* Evaluate the given slp. The optional arguments do the following : 
   var - Set to the variable wrt to which the DIFF operator nodes
	 are to be applied, if the Slp contains such nodes. 
   non_sing - Set to the address of an integer variable if you want
	 the evaluation procedure to determine if the Slp evaluated
	 to a singular value such as 0/0. 
   var2 - Set to the variable which has to be substituted for with
	  the next argument. That is when var2 is encountered in the
	  Slp, its value for evaluation will be 'fe' rather than
	  the random value which has been assigned to it. This
	  is useful when you want to evaluate the Slp at different
	  values of a variable such as when you want to interpolate
	  the Slp into a polynomial in that variable. 
   fe  -  Use in conjunction with var2, as explained above. Set
	  to the desired field element. 
*/

Field_element
evaluate(const Slp& slp, char* var = "",int* non_sing = 0,
	 char* var2 = "", Field_element fe = 0) 
{
  if (is_zero(slp)) 
    return(Field_element(0));
  else if (eval_level(slp) == Slp::Eval_level) {
    return(value(slp));
  }
  else if (op(slp) == NOP_FE) return(value(slp)); 
  else if (op(slp) == NOP_VAR) {
    if (!strcmp(var2,""))
      return(value(slp)); 
    else if (!strcmp(var2,name(variable_of(slp))))
      return(fe);
    else return(value(slp));
  }
  else if (op(slp) == DIFF) {
    /* If variable wrt which the derivative was taken is the same as 
       the Slp variable node to which the DIFF node points then the
       node represents the number 1 else it's 0. */
   if (is_variable(lchild(slp))) {
     if (!strcmp(var,name(variable_of(lchild(slp))))) 
       return Field_element(1); 
     else return Field_element(0);
   }
   else cout 
     << "Error in evaluate(Slp) - DIFF op used with non-variable operand\n";
  }
  else {
    Field_element result, lval, rval;
    if (non_sing != 0) {
      int l_non_sing = 1, r_non_sing = 1; 
      lval = evaluate(lchild(slp),var,&l_non_sing,var2,fe);  
      if (op(slp) != EXP) 
        rval = evaluate(rchild(slp),var,&r_non_sing,var2,fe);
      else r_non_sing = 1; 
      if (l_non_sing && r_non_sing) 
	*non_sing = 1; 
      else *non_sing = 0; 
    }
    else {
      lval = evaluate(lchild(slp),var,non_sing,var2,fe); 
      if (op(slp) != EXP) 
        rval = evaluate(rchild(slp),var,non_sing,var2,fe); 
    }
    if ((non_sing != 0) && (*non_sing == 0))
      return(Field_element(0));
    switch(op(slp)) {
      case ADD : 
	result = lval+rval;
        break;
      case SUB : 
	result = lval-rval;
        break;
      case MULT : 
	result = lval*rval;
        break;
      case DIV : 
	if (non_sing != 0) {
	  if (is_zero(rval))  
	    *non_sing = 0; 
	  else {
	    *non_sing = 1; 
	    result = lval/rval; 
	  }
	}
	else result = lval/rval;
        break;
      case EXP : 
	if (op(rchild(slp)) != NOP_FE) 
	  cout << "Error in Slp.evaluate()\n";
        else {
	  int temp;
	  rval = value(rchild(slp)); 
	  if (type_of(rval) != SMALLINT) 
	    cout << "Error in Slp.evaluate()\n";
	  else temp = ((SmallInt&)(rval)).i_value(); 
	  result = lval ^ temp;
	}
	break;
      default : cout << "Error in Slp.evaluate() - unknown operator\n"; 
    }
    slp.set_value(result);
    slp.eval_level(Slp::Eval_level);
    if (Debug) {
      cout << lval << "\n"; 
      cout << rval << "\n"; 
      cout << result << "\n"; 
    }
    return(result);
  }
}

Field_element
evaluate_modp(const Slp& slp, char* diff_var = "",int* non_sing = 0,
	 char* var = "", Field_element fe = 0) 
{
  if (strcmp(var,"")) { 
    if (non_sing != 0) { 
      return(evaluate_modp2a(slp,non_sing,var,fe,diff_var)); 
    }
    else { 
      return(evaluate_modp2b(slp,var,fe,diff_var)); 
    }
  }
  else { 
    if (non_sing != 0) { 
      return(evaluate_modp2c(slp,non_sing,diff_var)); 
    }
    else { 
      return(evaluate_modp2d(slp,diff_var)); 
    }
  }
}

Field_element
evaluate_modp2a(const Slp& slp, int* non_sing, char* var, Field_element fe,
		char* diff_var = "") 
{
  if (is_zero(slp)) 
    return(Field_element((unsigned)(0)));
  if (eval_level(slp) == Slp::Eval_level) {
    return(value(slp)); 
  }
  int slp_op = op(slp); 
  if (slp_op == NOP_FE) {
    if (type_of(value(slp)) == INTMODP)  
      return(value(slp)); 
    if (type_of(value(slp)) == SMALLINT)  
      return(make_Field_element((value(slp)).i_value(),INTMODP)); 
    else if (type_of(value(slp)) == BIGINT)  
      return(value(slp) * IntModp(1));
    else return(value(slp)); 
  }
  else if (slp_op == NOP_VAR) {
    if (!strcmp(var,name(variable_of(slp))))
      return(fe);
    else {
      // return(value(slp)); 
      if (type_of(value(slp)) == SMALLINT)  
        return(make_Field_element((value(slp)).i_value(),INTMODP)); 
      else return(value(slp)); 
    }
  }
  else if (slp_op == DIFF) {
    /* If variable wrt which the derivative was taken is the same as 
       the Slp variable node to which the DIFF node points then the
       node represents the number 1 else it's 0. */
   if (is_variable(lchild(slp))) {
     if (!strcmp(diff_var,name(variable_of(lchild(slp))))) 
       return Field_element((unsigned)1); 
     else return Field_element((unsigned)0);
   }
   else cout 
     << "Error in evaluate_modp2a(Slp) - DIFF op used with non-variable operand\n";
  }
  else {
    Field_element result, lval, rval;
    int l_non_sing = 1, r_non_sing = 1; 
    lval = evaluate_modp2a(lchild(slp),&l_non_sing,var,fe,diff_var);  
    if (slp_op != EXP) 
      rval = evaluate_modp2a(rchild(slp),&r_non_sing,var,fe,diff_var);
    else r_non_sing = 1; 
    if (l_non_sing && r_non_sing) *non_sing = 1; 
    else *non_sing = 0; 
    if (*non_sing == 0) { 
      result = Field_element((unsigned)0); 
      slp.set_value(result);
      slp.eval_level(Slp::Eval_level);
      return(result); 
    } 
    switch(slp_op) {
      case ADD : 
	result = lval+rval; break;
      case SUB : 
	result = lval-rval; break;
      case MULT : 
	result = lval*rval; break;
      case DIV : 
	if (is_zero(rval)) {
	    *non_sing = 0; 
	    result = (unsigned) 0; 
	}
	else {
	    *non_sing = 1; 
	    result = lval/rval; 
	}
        break;
      case EXP : 
	if (op(rchild(slp)) != NOP_FE) 
	  cout << "Error in Slp.evaluate_modp2a()\n";
        else {
	  int temp;
	  rval = value(rchild(slp)); 
	  if (type_of(rval) != SMALLINT) 
	    cout << "Error in Slp.evaluate_modp2a()\n";
	  else temp = ((SmallInt&)(rval)).i_value(); 
	  result = lval ^ temp;
	}
	break;
      default : cout << "Error in Slp.evaluate_modp2a() - unknown operator\n"; 
    }
    slp.set_value(result);
    slp.eval_level(Slp::Eval_level);
    if (Debug) 
      cout << result << "\n"; 
    return(result);
  }
}

Field_element
evaluate_modp2b(const Slp& slp, char* var, Field_element fe, 
		char* diff_var = "") 
{
  if (is_zero(slp)) 
    return(Field_element((unsigned)(0)));
  if (eval_level(slp) == Slp::Eval_level) {
    return(value(slp)); 
  }
  int slp_op = op(slp); 
  if (slp_op == NOP_FE) {
    if (type_of(value(slp)) == INTMODP)  
      return(value(slp)); 
    if (type_of(value(slp)) == SMALLINT)  
      return(make_Field_element((value(slp)).i_value(),INTMODP)); 
    else if (type_of(value(slp)) == BIGINT)  
      return(value(slp) * IntModp(1));
    else return(value(slp)); 
  }
  else if (slp_op == NOP_VAR) {
    if (!strcmp(var,name(variable_of(slp))))
      return(fe);
    else {
      // return(value(slp)); 
      if (type_of(value(slp)) == SMALLINT)  
        return(make_Field_element((value(slp)).i_value(),INTMODP)); 
      else return(value(slp)); 
    }
  }
  else if (slp_op == DIFF) {
    /* If variable wrt which the derivative was taken is the same as 
       the Slp variable node to which the DIFF node points then the
       node represents the number 1 else it's 0. */
   if (is_variable(lchild(slp))) {
     if (!strcmp(diff_var,name(variable_of(lchild(slp))))) 
       return Field_element((unsigned)1); 
     else return Field_element((unsigned)0);
   }
   else cout 
     << "Error in evaluate_modp2b(Slp) - DIFF op used with non-variable operand\n";
  }
  else {
    Field_element result, lval, rval;
    lval = evaluate_modp2b(lchild(slp),var,fe,diff_var); 
    if (slp_op != EXP) 
        rval = evaluate_modp2b(rchild(slp),var,fe,diff_var); 
    switch(slp_op) {
      case ADD : 
	result = lval+rval; break;
      case SUB : 
	result = lval-rval; break;
      case MULT : 
	result = lval*rval; break;
      case DIV : 
	result = lval/rval;
        break;
      case EXP : 
	if (op(rchild(slp)) != NOP_FE) 
	  cout << "Error in Slp.evaluate_modp2b()\n";
        else {
	  int temp;
	  rval = value(rchild(slp)); 
	  if (type_of(rval) != SMALLINT) 
	    cout << "Error in Slp.evaluate_modp2b()\n";
	  else temp = ((SmallInt&)(rval)).i_value(); 
	  result = lval ^ temp;
	}
	break;
      default : cout << "Error in Slp.evaluate_modp2b() - unknown operator\n"; 
    }
    slp.set_value(result);
    slp.eval_level(Slp::Eval_level);
    if (Debug) 
      cout << result << "\n"; 
    return(result);
  }
}

Field_element
evaluate_modp2c(const Slp& slp, int* non_sing,char* diff_var = "")
{
  if (is_zero(slp)) 
    return(Field_element((unsigned)(0)));
  if (eval_level(slp) == Slp::Eval_level) {
    return(value(slp)); 
  }
  int slp_op = op(slp); 
  if (slp_op == NOP_FE) {
    if (type_of(value(slp)) == INTMODP)  
      return(value(slp)); 
    if (type_of(value(slp)) == SMALLINT)  
      return(make_Field_element((value(slp)).i_value(),INTMODP)); 
    else if (type_of(value(slp)) == BIGINT)  
      return(value(slp) * IntModp(1));
    else return(value(slp)); 
  }
  else if (slp_op == NOP_VAR) {
      // return(value(slp)); 
      if (type_of(value(slp)) == SMALLINT)  
        return(make_Field_element((value(slp)).i_value(),INTMODP)); 
      else return(value(slp)); 
  }
  else if (slp_op == DIFF) {
    /* If variable wrt which the derivative was taken is the same as 
       the Slp variable node to which the DIFF node points then the
       node represents the number 1 else it's 0. */
   if (is_variable(lchild(slp))) {
     if (!strcmp(diff_var,name(variable_of(lchild(slp))))) 
       return Field_element((unsigned)1); 
     else return Field_element((unsigned)0);
   }
   else cout 
     << "Error in evaluate_modp2c(Slp) - DIFF op used with non-variable operand\n";
  }
  else {
    Field_element result, lval, rval;
    int l_non_sing = 1, r_non_sing = 1; 
    lval = evaluate_modp2c(lchild(slp),&l_non_sing,diff_var);  
    if (slp_op != EXP) 
      rval = evaluate_modp2c(rchild(slp),&r_non_sing,diff_var);
    else r_non_sing = 1; 
    if (l_non_sing && r_non_sing) *non_sing = 1; 
    else *non_sing = 0; 
    if (*non_sing == 0) { 
      result = Field_element((unsigned)0); 
      slp.set_value(result);
      slp.eval_level(Slp::Eval_level);
      return(result); 
    } 
    switch(slp_op) {
      case ADD : 
	result = lval+rval; break;
      case SUB : 
	result = lval-rval; break;
      case MULT : 
	result = lval*rval; break;
      case DIV : 
	if (is_zero(rval)) {
	    *non_sing = 0; 
	    result = (unsigned) 0; 
	}
	else {
	    *non_sing = 1; 
	    result = lval/rval; 
	}
        break;
      case EXP : 
	if (op(rchild(slp)) != NOP_FE) 
	  cout << "Error in Slp.evaluate_modp2c()\n";
        else {
	  int temp;
	  rval = value(rchild(slp)); 
	  if (type_of(rval) != SMALLINT) 
	    cout << "Error in Slp.evaluate_modp2c()\n";
	  else temp = ((SmallInt&)(rval)).i_value(); 
	  result = lval ^ temp;
	}
	break;
      default : cout << "Error in Slp.evaluate_modp2c() - unknown operator\n"; 
    }
    slp.set_value(result);
    slp.eval_level(Slp::Eval_level);
    if (Debug) 
      cout << result << "\n"; 
    return(result);
  }
}

Field_element
evaluate_modp2d(const Slp& slp, char* diff_var = "")
{
  if (is_zero(slp)) 
    return(Field_element((unsigned)(0)));
  if (eval_level(slp) == Slp::Eval_level) {
    return(value(slp)); 
  }
  int slp_op = op(slp); 
  if (slp_op == NOP_FE) {
    if (type_of(value(slp)) == INTMODP)  
      return(value(slp)); 
    if (type_of(value(slp)) == SMALLINT)  
      return(make_Field_element((value(slp)).i_value(),INTMODP)); 
    else if (type_of(value(slp)) == BIGINT)  
      return(value(slp) * IntModp(1));
    else return(value(slp)); 
  }
  else if (slp_op == NOP_VAR) {
      // return(value(slp)); 
      if (type_of(value(slp)) == SMALLINT)  
        return(make_Field_element((value(slp)).i_value(),INTMODP)); 
      else return(value(slp)); 
  }
  else if (slp_op == DIFF) {
    /* If variable wrt which the derivative was taken is the same as 
       the Slp variable node to which the DIFF node points then the
       node represents the number 1 else it's 0. */
   if (is_variable(lchild(slp))) {
     if (!strcmp(diff_var,name(variable_of(lchild(slp))))) 
       return Field_element((unsigned)1); 
     else return Field_element((unsigned)0);
   }
   else cout 
     << "Error in evaluate_modp2d(Slp) - DIFF op used with non-variable operand\n";
  }
  else {
    Field_element result, lval, rval;
    lval = evaluate_modp2d(lchild(slp),diff_var); 
    if (slp_op != EXP) 
        rval = evaluate_modp2d(rchild(slp),diff_var); 
    switch(slp_op) {
      case ADD : 
	result = lval+rval; break;
      case SUB : 
	result = lval-rval; break;
      case MULT : 
	result = lval*rval; break;
      case DIV : 
	result = lval/rval; break;
      case EXP : 
	if (op(rchild(slp)) != NOP_FE) 
	  cout << "Error in Slp.evaluate_modp2d()\n";
        else {
	  int temp;
	  rval = value(rchild(slp)); 
	  if (type_of(rval) != SMALLINT) 
	    cout << "Error in Slp.evaluate_modp2d()\n";
	  else temp = ((SmallInt&)(rval)).i_value(); 
	  result = lval ^ temp;
	}
	break;
      default : cout << "Error in Slp.evaluate_modp2d() - unknown operator\n"; 
    }
    slp.set_value(result);
    slp.eval_level(Slp::Eval_level);
    if (Debug) 
      cout << result << "\n"; 
    return(result);
  }
}

