
/*************************************************************************

  APU R 2.0

  slp_interp2.c

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  
#include <sys/types.h>
#include "slp.h"
#include "polynomial.h"
#include "poly_interp.h"
#include "crt.h"
#include "primelist.h"
extern int Debug; 

/* This routine is used for interpolating polynomials with BigInt
   coefficients. Vandermonde interpolation will not work since
   that requires division and will therefore require rational
   numbers. Instead, we use the Chinese remainder theorem.  

   The MultiIntModp field element type is used to evaluate the SLP
   to reduce the overhead of evaluation. This type allows the simultaneous
   evaluation of the SLP mod several primes (number given by the variable 
   MultiIntModp::CurrentNumPrimes). The variable "crt_yes" determines whether
   the polynomial coefficients, obtained in MultiIntModp form, should
   be interpolated into BigInt's using Chinese remaindering. 
*/

Polynomial
interpolate_multiint(const Slp& slp, char* var, bool slp_form = false,
bool crt_yes = false) 
{
  Polynomial result;
  if (is_zero(slp))
    result = Polynomial::zero_poly; 
  else {
    Variable* temp_v = Slp::var_list.find(var); 
    if (temp_v == 0) {
      cout << "Illegal variable in Slp - " << var << "\n";
      return(Polynomial::zero_poly); 
    }
    Slp::Degree_eval_level++; 
    int degree = degree_in_var(slp,var); 
    cout << "Degree of interpolating polynomial = " << degree << "\n"; 

    // This is the set up for efficient memory managing for the 
    // vanderpolate_mem() routine
    Array<Polynomial> poly_array = Array<Polynomial>(degree+1); 
    Array<Field_element> eval_points = Array<Field_element>(degree+1);
    Matrix<int> poly_modp_coeffs = Matrix<int>(degree+1,Primes.num_primes());  
    Array<Polynomial> master_coeffs = Array<Polynomial>(degree+2); 
    Array<Field_element> mp_coeffs = Array<Field_element>(degree+2); 
    Polynomial master_poly; 
    Polynomial Qi_sym = get_sym_quotient_poly(var,degree,MULTI_INTMODP); 
    Array<Field_element> coeff_array = Array<Field_element>(degree+1); 

    // This is the set up for efficient memory managing for the 
    // Chinese remainders routine. The array poly_coeffs stores the 
    // partial sums computed in Chinese remaindering. The array 
    // v has the "v" results in the algorithm in Knuth, v.2, p.274;
    // done_crt keeps track of which numbers are fully interpolated so far 
    // and the variable howmany_done tells how many these are, 
    // curr_primeprod is the product of the primes used till the current  
    // point; prime_index is the index of the prime used currently.  
    Array<BigInt> poly_coeffs = Array<BigInt>(degree+1); 
    Matrix<int> v = Matrix<int>(degree+1,Primes.num_primes()); 
    Array<bool> done_crt = Array<bool>(degree+1); 
    Array<BigInt> partial_sums = Array<BigInt>(degree+1); 
    int howmany_done = 0; 
    BigInt partial_sum,range,curr_primeprod, crt_result; 
    int prime_index = 0; 

    for (int j=0;j<degree+1;j++) 
      done_crt[j] = false; 

    int Done = 0; 
    int all_correct = 1; 
    int old_prime = IntModp::prime; 
    while (!Done) { /* Main loop which exits when Chinese remaindering is
		       over (if crt_yes = true) else when the polynomial
		       has gone through one pass of Vand. interpolation */
      if (prime_index % MultiIntModp::CurrentNumPrimes == 0) { 
        for (int i=0;i<MultiIntModp::CurrentNumPrimes;i++) { 
           MultiIntModp::primes[i] = Primes.primes[prime_index+i];
        }
        int extra_i = degree+2; 
        for (int i=1;i<=degree+1;i++) {
          Field_element fe = make_Field_element(i,MULTI_INTMODP); 
          Slp::Eval_level++; int done = 1; 
          Field_element fe_val; int non_sing; 
          fe_val = evaluate_multimodp(slp,"",&non_sing,var,fe); 
          if (Debug) {
	    cout << i << "\n"; 
	    cout << fe_val << "\n"; 
	  }
          if (!non_sing) 
	    done = 0; 
          while (!done) { 
	    // If the Slp is singular when evaluated at the given point
	    // we don't bother with differentiation using L'Hopital's rule -
	    // instead we just try another number. 
	    fe = make_Field_element(extra_i,MULTI_INTMODP); 
	    Slp::Eval_level++; 
	    fe_val = evaluate_multimodp(slp,"",&non_sing,var,fe); 
	    if (non_sing) done = 1; 
            if (Debug) { 
	      cout << fe_val << "\n"; 
	      cout << non_sing << "\n"; 
            } 
	    extra_i++;
          }
          if (Debug) 
	    cout << fe_val << "\n"; 
          eval_points[i-1] = fe; 
          poly_array[i-1] = Polynomial(fe_val); 
        }

        // Allocates memory for the interpolation polynomials etc. These
        // are then re-used intead of being allocated anew. 
        get_mp_coeffs_mem(mp_coeffs,degree,MULTI_INTMODP,eval_points); 
        get_masterpoly_coeffs_mem(master_coeffs,degree,mp_coeffs); 
        get_master_poly_mem(&master_poly,master_coeffs,degree,var); 


        result = vanderpolate_mem2(poly_array,eval_points,degree,var,
	  &master_poly,&Qi_sym,coeff_array,MULTI_INTMODP); 
      }
      if (crt_yes) { 
        TermList tl = termlist_of(result); 
        TermNode* tn;
        curr_primeprod = Primes.prime_prod(prime_index); 
        range = curr_primeprod/BigInt(2); 
        tn = tl.tnode_of(); 

        for (int j=degree;j>=0;j--) { 
          if ((tn == 0) || (tn->degree_of() != j)) { 
	      poly_modp_coeffs[j][prime_index] = 0; 
          }
          else {
	    if (!is_terminal(tn->coeff_poly())) { 
	      cout << "Error in interpolate_int() - non terminal poly\n"; 
	      return Polynomial::zero_poly; 
	    }
	    else {
	      poly_modp_coeffs[tn->degree_of()][prime_index] = 
	      (fe_of(tn->coeff_poly())).i_value(prime_index % MultiIntModp::CurrentNumPrimes); 
	      tn = tn->next_of(); 
	    }
	  }
	  if (!done_crt[j]) { 
	    partial_sum = crt(poly_modp_coeffs[j], 
	      v[j], prime_index, partial_sums[j]); 
	    if (range <= partial_sum) 
	      crt_result = partial_sum - curr_primeprod; 
            else crt_result = partial_sum; 
	    if (Debug) { 
	      cout << "Degree = " << j << "\n"; 
	      cout << MultiIntModp::primes[prime_index] << "\n"; 
	      cout << Primes.primes[prime_index] << "\n"; 
	      cout << poly_modp_coeffs[j][prime_index] << "\n"; 
	      cout << poly_coeffs[j] << "\n"; 
	      cout << crt_result << "\n"; 
	    }
	    if (crt_result == poly_coeffs[j]) {
	      done_crt[j] = true; 
	      howmany_done++; 
	    }
	    poly_coeffs[j] = crt_result; 
	    partial_sums[j] = partial_sum; 
          } 
        }
        if (howmany_done == degree+1)
	  Done = 1; 
        prime_index++; 
	if (Debug) { 
	  cout << prime_index << "\n"; 
	}
        if (prime_index >= Primes.num_primes()) { 
	  cout << "Primes exhausted - add more primes in file crt.c\n";
	  return(Polynomial::zero_poly); 
        }
      }
      else Done = 1; 
    }
    if (crt_yes) { 
      int i; 
      for (i=degree;i>=0;i--) {
        if (poly_coeffs[i] != 0) 
	  break; 
      }
      if (i < 0) result = Polynomial::zero_poly; 
      else { 
        TermList tl;
        if (slp_form)
          tl = TermList(TermNode(i,Polynomial(Slp(poly_coeffs[i])))); 
        else
          tl = TermList(TermNode(i,Polynomial(poly_coeffs[i]))); 
        result = Polynomial(var,tl); 
        TermNode* tn = tl.last(); 
        for (int j=i-1;j>=0;j--) { 
	  if (poly_coeffs[j] != 0) {
	    if (slp_form)
	      tn->set_next(TermNode(j,Polynomial(Slp(poly_coeffs[j]))));
	    else 
	      tn->set_next(TermNode(j,Polynomial(poly_coeffs[j])));
	    tn = tn->next_of(); 
	  }
        }
      }
      IntModp::prime = old_prime;
      cout << "Primes used = " << prime_index << "\n"; 
    }
    else { /* we're not using crt, however, slp_form could be true */
      if (slp_form) { 
        TermList tl = termlist_of(result);
        TermNode* tn = tl.tnode_of(); 
	for (tn=tl.tnode_of();tn!=0;tn=tn->next_of()) { 
          if (!is_terminal(tn->coeff_poly())) { 
	    cout << "Error in interpolate_int() - non terminal poly\n"; 
	    return Polynomial::zero_poly; 
          }
          else {
	    /* Make the coefficient of the coeff_poly of current term node
	     be an Slp of the original coefficient */
	    tn->set_coeff_poly(Polynomial(Slp(fe_of(tn->coeff_poly()))));
          }
	}
      }
    }
  }
  return result;
}

