
/*************************************************************************

  APU R 2.0

  sparse_res.c

  (c) 1996, Regents of the University of California. All rights reserved. 

  Author : Ashutosh Rege (rege@cs.berkeley.edu)

*************************************************************************/
  
#include "sparse_res.h"
#include "sparse_mat.h"
#include "molynomial.h"
extern int Debug; 

/* Computes the resultant matrix of the array of polys by calling the 
   sparse resultant program. The optional variable "mat_computed" tells
   it whether the matrix has already been computed and put in the file
   "mat_nospec" by the "far" program (the indices go in "aputemp.ind"). 
   If so, "far" will not be called and the matrix will be read directly 
   from "mat_nospec". The optional variable "metis" is to be set
   true if the package Metis is to be used to process the sparse
   resultant matrix to make it more amenable to computing the determinant */ 
   
Matrix<Slp> 
res_matrix(Array<Polynomial> polys, int num_vars, bool mat_computed = false,
int metis=0) 
{ 
  int n = size_of(polys); 
  Array<Molynomial> molys = Array<Molynomial>(n); 
  for (int i=0;i<n;i++) {
    molys[i] = poly_to_moly(polys[i]); 
    if (Debug) { 
      cout << molys[i] << "\n";
    }
  }
  return res_matrix(molys,num_vars,mat_computed,metis); 
}

Matrix<Slp>
res_matrix(Array<Molynomial> molys, int num_vars, bool mat_computed = false,
int metis=false) 
{ 
  int num_polys = size_of(molys); 
  if (num_polys < num_vars) { 
    cout << "Error in res_matrix(): Too few polynomials\n"; 
  }
  int max_monoms = compute_maxmonoms(molys); 
  Array<SlpArray> coeffs = compute_coeffs(molys); 
  if (Debug) { 
    Slp::Eval_level++; 
    Field_element fe; 
    Field_element fe2 = Field_element(2); 
    cout << fe2 << "\n"; 
    for (int i=0;i<size_of(coeffs);i++) { 
      for (int j=0;j<size_of(coeffs[i]);j++) { 
	if (type_of(coeffs[i][j]) == SLP) { 
	  int non_sing; 
	  fe = evaluate_modp(coeffs[i][j],"",&non_sing,"s",fe2); 
	  cout << fe << "  "; 
	}
	else cout << coeffs[i][j] << " "; 
      }
      cout << "\n"; 
    }
  }
  if (!mat_computed) { 
    Array<int> nfold_vols = nfold_mixvols(molys); 
    dump_for_far(num_vars,molys,nfold_vols); 
    system("./far -iw aputemp.ind -ms 0 -nco -pM -sy aputemp aputemp aputemp 2> aputemp.out >> aputemp.out"); 

  }
  /* At this point we have called the sparse res program and it has
     dumped into the file "mat_nospec" the matrix and the in the file
     "aputemp.ind" the row indices which correspond to a maximal minor,
     i.e. the resultant matrix. So now we read in the minor */

  Matrix<Slp> result; 
  if (metis) 
    result = create_matrix("mat_nospec","aputemp.ind",
	       num_polys,200,num_polys-1,coeffs,true,"aputemp.metis"); 
  else
    result = create_matrix("mat_nospec","aputemp.ind",num_polys,200,
	     num_polys-1,coeffs); 

  return result; 
}

Array<int>
nfold_mixvols(Array<Molynomial> molys)
{ 
  int n = size_of(molys); 
  Array<int> result = Array<int>(n); 
  Array<Molynomial> temp = Array<Molynomial>(n-1); 

  for (int i=0;i<n-1;i++) 
    temp[i] = molys[i+1]; 

  result[0] = mixvol(temp); 
  for (int i=1;i<n;i++) { 
    temp[i-1] = molys[i-1]; 
    result[i] = mixvol(temp); 
  }
  return result; 
}

void 
dump_for_far(int num_vars, Array<Molynomial> molys, Array<int> nfold_vols)
{
  char fname_prefix[8] = "aputemp"; 
  dump_sites(num_vars,molys,fname_prefix); 
  dump_vols(num_vars,nfold_vols,fname_prefix); 
}

void
dump_sites(int num_vars, Array<Molynomial> molys, char* prefix)
{
  int name_size = strlen(prefix) + 6; 
  char* fname = new char[name_size]; 
  strcpy(fname,prefix); 
  strcat(fname,".exps"); 
  ofstream file; 
  open_file(file,fname); 
  file << num_vars << "\n"; 
  file << size_of(molys) << "\n\n"; 
  int cur_count; 
  for (int i=0;i<size_of(molys);i++) { 
    cur_count = num_monoms(molys[i]); 
    file << cur_count << "\n"; 
  }
  file << "\n"; 
  for (int i=0;i<size_of(molys);i++) { 
    List<Monomial> cur_monoms = monoms_of(molys[i]); 
    ListNode<Monomial>* temp2; 
    for (temp2=head_of(cur_monoms);temp2!=0;temp2=temp2->next_of()) { 
      file << expons_of(temp2->item_of());
    }
    file << "\n"; 
  }
  file.close(); 
}

void
dump_vols(int num_vars, Array<int> vols, char* prefix) 
{ 
  int name_size = strlen(prefix) + 6; 
  char* fname = new char[name_size]; 
  strcpy(fname,prefix); 
  strcat(fname,".coef"); 
  ofstream file; 
  open_file(file,fname); 
  file << vols; 

  Array<int> vvec = Array<int>(num_vars); 
  for (int i=0;i<num_vars;i++) 
    vvec[i] = 50 + (int)(lrand48() % 100);

  file << vvec; 
  file.close(); 
}

int
compute_maxmonoms(Array<Molynomial> molys)
{
  int count = size_of(molys); 
  int result = 0; 
  int temp; 
  for (int i=0;i<count;i++) { 
    temp = num_monoms(molys[i]); 
    if (temp > result)
      result = temp; 
  }
  return result;
}

Array<SlpArray> 
compute_coeffs(Array<Molynomial> molys) 
{
  int dim1 = size_of(molys); 
  int temp; 
  Array<Field_element> fe_temp; 
  Array<SlpArray> result = Array<SlpArray>(dim1); 
  for (int i=0;i<dim1;i++) { 
    temp = num_monoms(molys[i]); 
    fe_temp = coeffs_of(molys[i]); 
    result[i] = Array<Slp>(size_of(fe_temp)); 
    for (int j=0;j<size_of(fe_temp);j++) { 
      if (type_of(fe_temp[j]) == SLP) 
        result[i][j] = (Slp&)fe_temp[j]; 
      else { 
	result[i][j] = Slp(fe_temp[j]); 
	if (Debug) 
	  cout << result[i][j] << "\n"; 
      }
    }
  }
  return result; 
}
